/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#ifndef _GRATINGMODELCMT_H
#define _GRATINGMODELCMT_H

#include <complex>
#include <vector>

#include "algorithms.h"
#include "sectionSolver.h"
#include "matrix_template.h"
#include "section.h"
#include "mode.h"

template<class modeModel, class numT>
class gratingModelCMT : public sectionSolver<numT>
{
 	private:
		section* f_sectionPtr;
 		std::vector<mode<modeModel, numT> > f_mode_vector;
		matrix<std::complex<numT> >* f_zeta_matrixPtr;
		numT f_normalisation;
		numT f_scaling;
		numT f_lambda;

	public:
		enum CMTMode
		{
			kCMTMODE_FULL,
			kCMTMODE_REDUCED
		};

	private:
		void set_zeta_matrix();
	
	public:
		gratingModelCMT(const std::vector<mode<modeModel, numT> >& mode_vector, const section& theSection, const numT& lambda, const numT& scaling = static_cast<numT>(1e6), const numT& normalisation = static_cast<numT>(1.0));
		gratingModelCMT(const gratingModelCMT<modeModel, numT>& object);
		~gratingModelCMT();
		static matrix<std::complex<numT> > jacobian(const numT& x, void* param);
		numT get_scaling();
		numT get_lambda();
		section* get_section_ptr();
		std::vector<mode<modeModel, numT> >* get_mode_vector_ptr();
		std::complex<numT> get_zeta_matrix_element(const unsigned int& row, const unsigned int& column);
		matrix<std::complex<numT> > amplitudes(const matrix<std::complex<numT> >& init);
};

template<class modeModel, class numT>
void gratingModelCMT<modeModel, numT>::set_zeta_matrix()
{
	layerDescription<numT> coreLayer = f_mode_vector[0].get_layer_description_for_radius(static_cast<numT>(0.0));
	numT k0Scaled_double = static_cast<numT>(2.0 * M_PI) / (f_lambda * f_scaling);
	
	for (int counter = 0; counter < f_mode_vector.size(); counter++)
	{
		f_mode_vector[counter].set_scaling(f_scaling);
		f_mode_vector[counter].set_normalisation(f_normalisation);
	}
	
	if (f_zeta_matrixPtr != 0)
		delete f_zeta_matrixPtr;

	f_zeta_matrixPtr = new matrix<std::complex<numT> >(f_mode_vector.size(), f_mode_vector.size());

	for (unsigned int rowCounter_uint = 0; rowCounter_uint < f_mode_vector.size(); rowCounter_uint++)
		for (unsigned int columnCounter_uint = 0; columnCounter_uint <= rowCounter_uint; columnCounter_uint++)
		{
			numT azimuthalIntegral = static_cast<numT>(2.0 * M_PI);
			std::complex<numT> radialIntegral = coupling_integral_radial(f_mode_vector[rowCounter_uint], f_mode_vector[columnCounter_uint], static_cast<numT>(0.0), coreLayer.r);
	 		(*f_zeta_matrixPtr)(rowCounter_uint, columnCounter_uint) = (k0Scaled_double * static_cast<std::complex<numT> >(std::sqrt(kEPSILON_0 / kMU_0)) / (f_scaling * f_scaling)) / (static_cast<numT>(2.0) * f_normalisation) * coreLayer.n * static_cast<numT>(f_sectionPtr->get_p_change()) * azimuthalIntegral * radialIntegral;

			if (rowCounter_uint != columnCounter_uint)
				(*f_zeta_matrixPtr)(columnCounter_uint, rowCounter_uint) = std::conj((*f_zeta_matrixPtr)(rowCounter_uint, columnCounter_uint));
		}
}

template<class modeModel, class numT>
matrix<std::complex<numT> > gratingModelCMT<modeModel, numT>::jacobian(const numT& x, void* param)
{
	gratingModelCMT* modeSolver = (gratingModelCMT*)param;
	
	std::vector<mode<modeModel, numT> >* mode_vectorPtr = modeSolver->get_mode_vector_ptr();
	numT k0Scaled = static_cast<numT>(2.0 * M_PI) / (modeSolver->get_lambda() * modeSolver->get_scaling());
	numT sigma = static_cast<numT>(modeSolver->get_section_ptr()->get_sigma(doubleValue(x / modeSolver->get_scaling())));

	matrix<std::complex<numT> > fy(mode_vectorPtr->size(), mode_vectorPtr->size());
	
	for (int counter1 = 0; counter1 < mode_vectorPtr->size(); counter1++)
	{
		for (int counter2 = 0; counter2 < mode_vectorPtr->size(); counter2++)
		{
			if (counter1 == counter2)
			{
				fy(counter1, counter2) = sigma * std::complex<numT>(static_cast<numT>(0.0), -static_cast<numT>(modeSolver->get_section_ptr()->get_dc_change())) * modeSolver->get_zeta_matrix_element(counter1, counter2) * std::exp(-2.0 * std::abs(k0Scaled * std::abs(std::imag((*mode_vectorPtr)[counter1].get_nEff())) * x));
			}
			else
			{
				numT accumulatedPhase_double = static_cast<numT>(modeSolver->get_section_ptr()->get_accumulated_phase(doubleValue(x / modeSolver->get_scaling())));
					
				std::complex<numT> exponent;
				if (std::real((*mode_vectorPtr)[counter1].get_nEff()) < std::real((*mode_vectorPtr)[counter2].get_nEff()))
				{
					// +
					exponent = std::complex<numT>(static_cast<numT>(0.0), x * k0Scaled * (std::real((*mode_vectorPtr)[counter1].get_nEff()) - std::real((*mode_vectorPtr)[counter2].get_nEff())) + accumulatedPhase_double);
				}
				else
				{
					// -
					exponent = std::complex<numT>(static_cast<numT>(0.0), x * k0Scaled * (std::real((*mode_vectorPtr)[counter1].get_nEff()) - std::real((*mode_vectorPtr)[counter2].get_nEff())) - accumulatedPhase_double);
				}
				fy(counter1, counter2) = sigma * std::complex<numT>(static_cast<numT>(0.0), -modeSolver->get_section_ptr()->get_ac_change() / 2.0) * modeSolver->get_zeta_matrix_element(counter1, counter2) * std::exp(exponent) * std::exp(-x * k0Scaled * std::abs(std::imag((*mode_vectorPtr)[counter1].get_nEff() + (*mode_vectorPtr)[counter2].get_nEff())));
			}
		}
	}
	
	return fy;
}

template <class modeModel, class numT>
gratingModelCMT<modeModel, numT>::gratingModelCMT(const std::vector<mode<modeModel, numT> >& mode_vector, const section& theSection, const numT& lambda, const numT& scaling, const numT& normalisation)
{
	f_mode_vector = mode_vector;
	f_sectionPtr = new section(theSection);
	f_zeta_matrixPtr = 0;
	f_lambda = lambda;
	f_scaling = scaling;
	f_normalisation = normalisation;
}

template <class modeModel, class numT>
gratingModelCMT<modeModel, numT>::gratingModelCMT(const gratingModelCMT<modeModel, numT>& object)
{
	f_mode_vector = object.f_mode_vector;
	f_sectionPtr = object._theGrating->clone();

	if (object._zeta_matrix != 0)
		f_zeta_matrixPtr = new matrix<std::complex<numT> >(*(object._zeta_matrix));
	else
		f_zeta_matrixPtr = 0;

	f_lambda = object.f_lambda;
	f_scaling = object.f_scaling;
	f_normalisation = object.f_normalisation;
}

template <class modeModel, class numT>
gratingModelCMT<modeModel, numT>::~gratingModelCMT()
{
	delete f_sectionPtr;
	delete f_zeta_matrixPtr;
}

template<class modeModel, class numT>
numT gratingModelCMT<modeModel, numT>::get_scaling()
{
	return f_scaling;
}

template<class modeModel, class numT>
numT gratingModelCMT<modeModel, numT>::get_lambda()
{
	return f_lambda;
}

template<class modeModel, class numT>
section* gratingModelCMT<modeModel, numT>::get_section_ptr()
{
	return f_sectionPtr;
}

template<class modeModel, class numT>
std::vector<mode<modeModel, numT> >* gratingModelCMT<modeModel, numT>::get_mode_vector_ptr()
{
	return &f_mode_vector;
}

template<class modeModel, class numT>
std::complex<numT> gratingModelCMT<modeModel, numT>::get_zeta_matrix_element(const unsigned int& row, const unsigned int& column)
{
	if (f_zeta_matrixPtr == 0) {
		set_zeta_matrix();
	}

	return (*f_zeta_matrixPtr)(row, column);
}

template<class modeModel, class numT>
matrix<std::complex<numT> > gratingModelCMT<modeModel, numT>::amplitudes(const matrix<std::complex<numT> >& init)
{
	return solve_ode_system(init, static_cast<numT>(0.0), static_cast<numT>(f_sectionPtr->get_length()), f_scaling, &(gratingModelCMT<modeModel, numT>::jacobian), this);
}

#endif

