/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#include "section.h"

section::section(const std::wstring& type) throw (std::runtime_error)
{
	if (type == L"")
		throw std::runtime_error("No type defined.");
	else if (type == L"grating")
		f_eos = section_grating::init();
		
	f_type = type;
}

section::section(const section& obj)
{
	f_eos = obj.f_eos;
	f_type = obj.f_type;
}

section::~section()
{
	return;
}

std::map<std::wstring, std::complex<double> > section::get_vars()
{
	std::map<std::wstring, std::complex<double> > vars;
	
	for (std::map<std::wstring, eo<std::complex<double> > >::iterator it = f_eos.begin(); it != f_eos.end(); it++)
	{
		if ((it->first != L"period") || (it->first != L"accumulated_phase") || (it->first != L"sigma"))
		{
			std::map<std::wstring, std::complex<double> > temp;
			it->second.get_vars(temp);
			temp.erase(L"position");
			vars.insert(temp.begin(), temp.end());
		}
		else
			it->second.get_vars(vars);
	}
	
	return vars;
}

void section::set_vars(const std::map<std::wstring, std::complex<double> >& vars)
{
	for (std::map<std::wstring, eo<std::complex<double> > >::iterator it = f_eos.begin(); it != f_eos.end(); it++)
		it->second.set_vars(vars);
}

void section::rename_vars(const std::map<std::wstring, std::wstring >& vars)
{
	for (std::map<std::wstring, eo<std::complex<double> > >::iterator it = f_eos.begin(); it != f_eos.end(); it++)
		it->second.rename_vars(vars);
}

void section::simplify()
{
	for (std::map<std::wstring, eo<std::complex<double> > >::iterator it = f_eos.begin(); it != f_eos.end(); it++)
		it->second.simplify();
}

double section::get_period(double position) const throw (std::runtime_error)
{
	if (f_eos.count(L"period") > 0)
	{
		eo<std::complex<double> > temp(f_eos.find(L"period")->second);
		std::map<std::wstring, std::complex<double> > pos;
		pos.insert(std::pair<std::wstring, std::complex<double> >(L"position", std::complex<double>(position)));
		
		temp.set_vars(pos);
		return std::real((std::complex<double>)temp);
	}
	
	std::stringstream error;
	error << "Function \"get_period\" was not defined in class \"" << wstos(f_type) << "\"";
	throw std::runtime_error(error.str());
}

double section::get_accumulated_phase(double position) const throw (std::runtime_error)
{
	if (f_eos.count(L"accumulated_phase") > 0)
	{
		eo<std::complex<double> > temp(f_eos.find(L"accumulated_phase")->second);
		std::map<std::wstring, std::complex<double> > pos;
		pos.insert(std::pair<std::wstring, std::complex<double> >(L"position", std::complex<double>(position)));
		
		temp.set_vars(pos);
		return std::real((std::complex<double>)temp);
	}
	
	std::stringstream error;
	error << "Function \"get_accumulated_phase\" was not defined in class \"" << wstos(f_type) << "\"";
	throw std::runtime_error(error.str());
}

double section::get_sigma(const double& position) const throw (std::runtime_error)
{
	if (f_eos.count(L"sigma") > 0)
	{
		eo<std::complex<double> > temp(f_eos.find(L"sigma")->second);
		std::map<std::wstring, std::complex<double> > pos;
		pos.insert(std::pair<std::wstring, std::complex<double> >(L"position", std::complex<double>(position)));
		
		temp.set_vars(pos);
		return std::real((std::complex<double>)temp);
	}
	
	std::stringstream error;
	error << "Function \"get_sigma\" was not defined in class \"" << wstos(f_type) << "\"";
	throw std::runtime_error(error.str());
}

double section::get_max_period(void) const throw (std::runtime_error)
{
	if (f_eos.count(L"max_period") > 0)
		return std::real((std::complex<double>)f_eos.find(L"max_period")->second);
	
	std::stringstream error;
	error << "Function \"get_max_period\" was not defined in class \"" << wstos(f_type) << "\"";
	throw std::runtime_error(error.str());
}

double section::get_min_period(void) const throw (std::runtime_error)
{
	if (f_eos.count(L"min_period") > 0)
		return std::real((std::complex<double>)f_eos.find(L"min_period")->second);
	
	std::stringstream error;
	error << "Function \"get_min_period\" was not defined in class \"" << wstos(f_type) << "\"";
	throw std::runtime_error(error.str());
}

double section::get_length(void) const throw (std::runtime_error)
{
	if (f_eos.count(L"length") > 0)
		return std::real((std::complex<double>)f_eos.find(L"length")->second);
	
	std::stringstream error;
	error << "Function \"get_length\" was not defined in class \"" << wstos(f_type) << "\"";
	throw std::runtime_error(error.str());
}

double section::get_dc_change(void) const throw (std::runtime_error)
{
	if (f_eos.count(L"dc_change") > 0)
		return std::real((std::complex<double>)f_eos.find(L"dc_change")->second);
	
	std::stringstream error;
	error << "Function \"get_dc_change\" was not defined in class \"" << wstos(f_type) << "\"";
	throw std::runtime_error(error.str());
}

double section::get_ac_change(void) const throw (std::runtime_error)
{
	if (f_eos.count(L"ac_change") > 0)
		return std::real((std::complex<double>)f_eos.find(L"ac_change")->second);
	
	std::stringstream error;
	error << "Function \"get_ac_change\" was not defined in class \"" << wstos(f_type) << "\"";
	throw std::runtime_error(error.str());
}

double section::get_p_change(void) const throw (std::runtime_error)
{
	if (f_eos.count(L"p_change") > 0)
		return std::real((std::complex<double>)f_eos.find(L"p_change")->second);
	
	std::stringstream error;
	error << "Function \"get_p_change\" was not defined in class \"" << wstos(f_type) << "\"";
	throw std::runtime_error(error.str());
}

std::wstring section::get_type(void) const
{
		return f_type;
}

std::map<std::wstring, eo<std::complex<double> > > section_grating::init()
{
	std::map<std::wstring, eo<std::complex<double> > > eos;
	
	eo<std::complex<double> > period(L"period");
	
	eos.insert(std::pair<std::wstring, eo<std::complex<double> > >(L"period", period));
	eos.insert(std::pair<std::wstring, eo<std::complex<double> > >(L"accumulated_phase", 2.0 * M_PI / period * eo<std::complex<double> >(L"position")));
	eos.insert(std::pair<std::wstring, eo<std::complex<double> > >(L"sigma", eo<std::complex<double> >(1.0)));
	eos.insert(std::pair<std::wstring, eo<std::complex<double> > >(L"max_period", period));
	eos.insert(std::pair<std::wstring, eo<std::complex<double> > >(L"min_period", period));
	eos.insert(std::pair<std::wstring, eo<std::complex<double> > >(L"length", eo<std::complex<double> >(L"length")));
	eos.insert(std::pair<std::wstring, eo<std::complex<double> > >(L"dc_change", eo<std::complex<double> >(L"acChange")));
	eos.insert(std::pair<std::wstring, eo<std::complex<double> > >(L"ac_change", eo<std::complex<double> >(L"dcChange")));
	eos.insert(std::pair<std::wstring, eo<std::complex<double> > >(L"p_change", eo<std::complex<double> >(L"pChange")));

	return eos;
}
