/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#include "XMLFile.h"

XMLString::XMLString(const std::wstring& str)
{
	size_t len = str.length();
	xmlChar buffer[len * 2 + 1];
	size_t resultLen = wcstombs((char*)buffer, str.c_str(), len);
	buffer[resultLen] = '\0';
	f_xmlCharPtr = xmlStrndup(buffer, resultLen + 1);
}

XMLString::XMLString(xmlChar* xmlCharPtr)
{
	f_xmlCharPtr = xmlCharPtr;
}

XMLString::XMLString(const XMLString& xmlStr)
{
	f_xmlCharPtr = xmlStrndup(xmlStr.f_xmlCharPtr, xmlStrlen(xmlStr.f_xmlCharPtr));
}

XMLString::operator std::wstring()
{
	size_t len = xmlStrlen(f_xmlCharPtr);
	wchar_t buffer[len];
	size_t resultLen = mbstowcs(buffer, (const char*)f_xmlCharPtr, len);
	return std::wstring(buffer, resultLen);
}

XMLString::operator xmlChar*()
{
	return f_xmlCharPtr;
}

XMLString::~XMLString()
{
	if (f_xmlCharPtr != 0)
		xmlFree(f_xmlCharPtr);
}

XMLNode::XMLNode()
{
	f_xmlNode = 0;
}

XMLNode::XMLNode(xmlNodePtr xmlNd)
{
	f_xmlNode = xmlNd;
}

XMLNode::XMLNode(const XMLNode& xmlNd)
{
	f_xmlNode = xmlNd.f_xmlNode;
}

XMLNode::~XMLNode()
{
}

std::wstring XMLNode::get_property(const std::wstring& prop) const
{
	if (f_xmlNode == 0)
		return std::wstring();

	xmlChar* temp = xmlGetProp(f_xmlNode, XMLString(prop));
	std::wstring output;
	
	if (temp != 0)
		output = XMLString(temp);
	
	return output;
}

std::wstring XMLNode::get_string() const
{
	if (f_xmlNode != 0)
		return XMLString(xmlNodeListGetString(f_xmlNode->doc, f_xmlNode->children, 0));
	else
		return std::wstring();
}

XPathObject::XPathObject(xmlXPathObjectPtr objPtr)
{
	f_xpathObj = objPtr;
}

XPathObject::XPathObject(const XPathObject& obj)
{
	f_xpathObj = xmlXPathObjectCopy(obj.f_xpathObj);
}

XPathObject::~XPathObject()
{
	xmlXPathFreeObject(f_xpathObj);
	f_xpathObj = 0;
}

int XPathObject::get_numberOfNodes() const
{
	if (isNodeSet(*this))
		if (f_xpathObj->nodesetval != 0)
			return f_xpathObj->nodesetval->nodeNr;
	return 0;
}

XMLNode XPathObject::get_node(const int& num) const
{
	if ((num >= 0) && (get_numberOfNodes() > num))
		return XMLNode(*(f_xpathObj->nodesetval->nodeTab + num));
	else
		return XMLNode();
}

bool isNodeSet(const XPathObject& obj)
{
	if ((obj.f_xpathObj != 0) && (obj.f_xpathObj->type == XPATH_NODESET))
		return true;
	else
		return false;
}

void XMLFile::initiateParserContext(void) throw (std::runtime_error)
{
	f_parserCtxt = xmlNewParserCtxt();
	if (f_parserCtxt == 0)
		throw std::runtime_error("Failed to allocate parser context");
}

void XMLFile::initiateXPathContext(void) throw (std::runtime_error)
{
	f_XPathCtxt = xmlXPathNewContext(f_doc);
	if (f_XPathCtxt == 0)
		throw std::runtime_error("Failed to allocate XPath context");
}

XMLFile::XMLFile(void)
{
	f_parserCtxt = 0;
	f_doc = 0;
	f_XPathCtxt = 0;
	initiateParserContext();
}

XMLFile::XMLFile(const std::wstring& filename)
{
	f_parserCtxt = 0;
	f_doc = 0;
	f_XPathCtxt = 0;
	initiateParserContext();
	open(filename);
}

XMLFile::~XMLFile(void)
{
	close(false);
	if (f_parserCtxt != NULL)
		xmlFreeParserCtxt(f_parserCtxt);
}

void XMLFile::open(const std::wstring& filename)
{
	if ((f_doc == 0) && (f_parserCtxt != 0))
	{
		f_doc = xmlCtxtReadFile(f_parserCtxt, wstos(filename).c_str(), 0, 0);
		if (f_doc == 0)
			throw std::runtime_error("Failed to parse file");

		initiateXPathContext();
	}
	else
		throw std::runtime_error("File previousily open must be closed first");
}

void XMLFile::close(bool throwExceptions) throw (std::runtime_error)
{
	if (f_XPathCtxt != 0)
	{
		xmlXPathFreeContext(f_XPathCtxt);
		f_XPathCtxt = 0;
	}
	
	if (f_doc != 0)
	{
		xmlFreeDoc(f_doc);
		f_doc = 0;
	}
}

void XMLFile::saveAs(const std::wstring& filename)
{
	if (f_doc != 0)
		xmlSaveFileEnc(wstos(filename).c_str(), f_doc, "UTF-8");
}

bool XMLFile::isOpen(void)
{
	if ((f_doc != 0) && (f_parserCtxt != 0))
		return true;
	else
		return false;
}

XPathObject XMLFile::XPathQuery(const std::wstring& query, const std::wstring& prefix, const std::wstring& href)
{
	std::map<std::wstring, std::wstring> map;
	map[prefix] = href;
	return XPathQuery(query, map);
}

XPathObject XMLFile::XPathQuery(const std::wstring& query, const std::map<std::wstring, std::wstring>& namespaces)
{
	xmlNs nsList[namespaces.size()];
	int counter = 0;

	// Not very elegant. Have to find a better way of doing this.
	for (std::map<std::wstring, std::wstring>::const_iterator it = namespaces.begin(); it != namespaces.end(); it++)
	{
		nsList[counter].prefix = xmlStrdup(XMLString(it->first));
		nsList[counter++].href = xmlStrdup(XMLString(it->second));
	}

	xmlNs* nsP = (xmlNs*)&nsList;
	
	f_XPathCtxt->namespaces = &nsP;
	f_XPathCtxt->nsNr = namespaces.size();

	XPathObject output(xmlXPathEvalExpression(XMLString(query), f_XPathCtxt));

	f_XPathCtxt->namespaces = 0;
	f_XPathCtxt->nsNr = 0;

	for (int counter = 0; counter < namespaces.size(); counter++)
	{
		xmlFree((xmlChar*)nsList[counter].prefix);
		nsList[counter].prefix = 0;
		xmlFree((xmlChar*)nsList[counter].href);
		nsList[counter].href = 0;
	}

	return output;
}

const xmlDocPtr XMLFile::getXMLDocPtr(void)
{
	if (f_doc != 0)
		return f_doc;
	else
		return 0;
}

