/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#ifndef _MOGRASOJOBMODESTRUCTURE_H
#define _MOGRASOJOBMODESTRUCTURE_H

#include <map>
#include <string>

#include "modeSolver.h"
#include "modeModelHE.h"
#include "modeModelLP.h"
#include "gratingModelCMT.h"

template <class modeModel, class numT>
class mograsoJobModeStructure : public jobInterface
{
	std::map<std::wstring, std::complex<double> > f_varSet;
	std::vector<layerDescription<numT> > f_layers;
	std::map<std::wstring, pointerWrapper<std::ofstream> > f_fileMap;
	std::wstring f_lineStart;
	std::map<std::wstring, std::wstring> f_options;
	
	public:
		mograsoJobModeStructure(const std::map<std::wstring, std::complex<double> >& varSet, const std::vector<layerDescription<numT> >& layers, const std::map<std::wstring, pointerWrapper<std::ofstream> >& fileMap, const std::wstring& lineStart, const std::map<std::wstring, std::wstring>& options);
		virtual ~mograsoJobModeStructure() {};
		virtual void job();
		virtual std::map<std::wstring, pointerWrapper<std::ofstream> >& get_fileMap();
		static std::vector<std::wstring> perJobFilenames();
		static std::vector<std::wstring> perVarSetFilenames();
	
	private:
		mograsoJobModeStructure();
		mograsoJobModeStructure(const mograsoJobModeStructure&);
};

template <class modeModel, class numT>
mograsoJobModeStructure<modeModel, numT>::mograsoJobModeStructure(const std::map<std::wstring, std::complex<double> >& varSet, const std::vector<layerDescription<numT> >& layers, const std::map<std::wstring, pointerWrapper<std::ofstream> >& fileMap, const std::wstring& lineStart, const std::map<std::wstring, std::wstring>& options)
{
	f_varSet = varSet;
	f_layers = layers;
	f_fileMap = fileMap;
	f_lineStart = lineStart;
	f_options = options;
}

template <class modeModel, class numT>
void mograsoJobModeStructure<modeModel, numT>::job()
{
	try
	{
		numT lambda = NAN;
		numT scaling = static_cast<numT>(1e6);
		
		if (f_fileMap.count(L"log") > 0)
		{
			pointerWrapper<std::ofstream> log = f_fileMap.find(L"log")->second;
			log.lock();
			*(log.getObjectPtr()) << PACKAGE_STRING << "\n\n";
			*(log.getObjectPtr()) << "The following options were chosen:\n";
			for (std::map<std::wstring, std::wstring>::iterator it = f_options.begin(); it != f_options.end(); it++)
				*(log.getObjectPtr()) << wstos(it->first) << ": " << wstos(it->second) << "\n";
			*(log.getObjectPtr()) << "\n";
			log.unlock();			
		}

		if (f_varSet.count(L"wavelength") > 0)
			lambda = std::abs(f_varSet[L"wavelength"]);
		else
			return;
		
		if (f_fileMap.count(L"log") > 0)
		{
			pointerWrapper<std::ofstream> log = f_fileMap.find(L"log")->second;
			log.lock();
			*(log.getObjectPtr()) << "Wavelength: " << doubleValue(lambda) << "\n\n";
			log.unlock();			
		}		
		
		if (f_fileMap.count(L"log") > 0)
		{
			pointerWrapper<std::ofstream> log = f_fileMap.find(L"log")->second;
			log.lock();
			*(log.getObjectPtr()) << "Layers:\n";
			for (typename std::vector<layerDescription<numT> >::const_iterator it = f_layers.begin(); it != f_layers.end(); it++)
				*(log.getObjectPtr()) << "r: " << doubleValue(it->r) << ", e: " << doubleValue(it->e) << ", u: " << doubleValue(it->u) << "\n";
			*(log.getObjectPtr()) << "\n";
			log.unlock();
		}
		
		modeSolver<modeModel, numT> model(f_layers, lambda, scaling, static_cast<numT>(1e-10), static_cast<numT>(1e-5), &f_fileMap);
		
		// Find the maximum and minimum real and imaginary refractive indices of the layers.
		numT realMaxN = std::real(f_layers[1].n);
		numT realMinN = std::real(f_layers[1].n) * static_cast<numT>(0.999);
		numT imagMaxN = static_cast<numT>(1e-4);
		numT imagMinN = static_cast<numT>(-1e-5);
		
		for (int counter = 0; counter < (f_layers.size() - 1); counter++)
		{
			realMaxN = std::max(realMaxN, std::real(f_layers[counter].n));
			imagMaxN = std::max(imagMaxN, std::imag(f_layers[counter].n));
			imagMinN = std::min(imagMinN, std::imag(f_layers[counter].n));
		}

		// Check user defined maxima and minima.
		if (f_varSet.count(L"realMaxN") > 0)
			realMaxN = std::real(f_varSet[L"realMaxN"]);

		if (f_varSet.count(L"realMinN") > 0)
			realMinN = std::real(f_varSet[L"realMinN"]);

		if (f_varSet.count(L"imagMaxN") > 0)
			imagMaxN = std::real(f_varSet[L"imagMaxN"]);

		if (f_varSet.count(L"imagMinN") > 0)
			imagMinN = std::real(f_varSet[L"imagMinN"]);		
		
		interval<std::complex<numT> > interval = {std::complex<numT>(realMaxN, imagMaxN), std::complex<numT>(realMinN, imagMinN)};

		if (f_fileMap.count(L"log") > 0)
		{
			pointerWrapper<std::ofstream> log = f_fileMap.find(L"log")->second;
			log.lock();
			*(log.getObjectPtr()) << "Mode interval: " << doubleValue(interval.start_T) << "-" << doubleValue(interval.end_T) << "\n\n";
			log.unlock();
		}

		mode<modeModel, numT> coreMode = model.determine_core_mode();
		std::vector<mode<modeModel, numT> > outputModes;
		outputModes.push_back(coreMode);
		std::vector<mode<modeModel, numT> > modes = model.determine_modes_in_interval(interval);
		
		std::list<mode<modeModel, numT> > modeList(modes.begin(), modes.end());
		modeList.sort();
		modeList.reverse();
		modes.clear();
		modes.insert(modes.begin(), modeList.begin(), modeList.end());
		
		for (int counter = 0; counter < modes.size(); counter++)
			if (!thresholdEqual(modes[counter].get_nEff(), coreMode.get_nEff(), static_cast<numT>(1e-15)))
				outputModes.push_back(modes[counter]);
				
		
		if (f_fileMap.count(L"modes") > 0)
		{
			pointerWrapper<std::ofstream> log = f_fileMap.find(L"modes")->second;
			log.lock();
			log.getObjectPtr()->precision(16);
			*(log.getObjectPtr()) << wstos(f_lineStart);
			for (int counter = 0; counter < outputModes.size(); counter++)
			{
				*(log.getObjectPtr()) << std::real(outputModes[counter].get_nEff()) << "\t" << std::imag(outputModes[counter].get_nEff()) << "\t";
				if (f_options.count(L"normalisedSelfCouplingCoefficient"))
				{
					numT azimuthalIntegral = static_cast<numT>(2.0 * M_PI);
					outputModes[counter].set_scaling(static_cast<numT>(1.0));
					outputModes[counter].set_normalisation(static_cast<numT>(1.0));
					std::complex<numT> radialIntegral = coupling_integral_radial(outputModes[counter], outputModes[counter], static_cast<numT>(0.0), f_layers[0].r);
	 				std::complex<numT> nSCC = static_cast<std::complex<numT> >(std::sqrt(kEPSILON_0 / kMU_0) / 2.0) * f_layers[0].n * azimuthalIntegral * radialIntegral;
					*(log.getObjectPtr()) << doubleValue(std::real(nSCC)) << "\t" << doubleValue(std::imag(nSCC)) << "\t";
				}
			}
			*(log.getObjectPtr()) << "\n";
			log.unlock();
		}
		if (f_fileMap.count(L"log") > 0)
		{
			pointerWrapper<std::ofstream> log = f_fileMap.find(L"log")->second;
			log.lock();
			log.getObjectPtr()->precision(16);
			*(log.getObjectPtr()) << "Modes found:\n";
			for (int counter = 0; counter < outputModes.size(); counter++)
				*(log.getObjectPtr()) << doubleValue(std::real(outputModes[counter].get_nEff())) << "\t" << doubleValue(std::imag(outputModes[counter].get_nEff())) << "\t";
			*(log.getObjectPtr()) << "\n";
			log.unlock();
		}
	}
	catch (std::runtime_error e)
	{
		if (f_fileMap.count(L"log") > 0)
		{
			pointerWrapper<std::ofstream> log = f_fileMap.find(L"log")->second;
			log.lock();
			*(log.getObjectPtr()) << e.what() << "\n";
			log.unlock();
		}		
	}
	catch (...)
	{
		if (f_fileMap.count(L"log") > 0)
		{
			pointerWrapper<std::ofstream> log = f_fileMap.find(L"log")->second;
			log.lock();
			*(log.getObjectPtr()) << "Error\n";
			log.unlock();
		}		
	}
}

template <class modeModel, class numT>
std::vector<std::wstring> mograsoJobModeStructure<modeModel, numT>::perJobFilenames()
{
	std::vector<std::wstring> out;
	out.push_back(L"modes");
	return out;
}

template <class modeModel, class numT>
std::vector<std::wstring> mograsoJobModeStructure<modeModel, numT>::perVarSetFilenames()
{
	std::vector<std::wstring> out;
	out.push_back(L"log");
	return out;
}

template <class modeModel, class numT>
std::map<std::wstring, pointerWrapper<std::ofstream> >& mograsoJobModeStructure<modeModel, numT>::get_fileMap()
{
	return f_fileMap;
}

#endif

