/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#ifndef _MOGRASOJOBSPECTRUM_H
#define _MOGRASOJOBSPECTRUM_H

#include <map>
#include <string>

#include "modeSolver.h"
#include "modeModelHE.h"
#include "modeModelLP.h"
#include "gratingModelCMT.h"
#include "gratingModelTM.h"

template <class modeModel, class numT>
class mograsoJobSpectrum : public jobInterface
{
	std::map<std::wstring, std::complex<double> > f_varSet;
	std::vector<layerDescription<numT> > f_layers;
	std::vector<section> f_sections;
	std::map<std::wstring, pointerWrapper<std::ofstream> > f_fileMap;
	std::wstring f_lineStart;
	std::map<std::wstring, std::wstring> f_options;
	
	public:
		mograsoJobSpectrum(const std::map<std::wstring, std::complex<double> >& varSet, const std::vector<layerDescription<numT> >& layers, const std::vector<section>& sections, const std::map<std::wstring, pointerWrapper<std::ofstream> >& fileMap, const std::wstring& lineStart, const std::map<std::wstring, std::wstring>& options);
		virtual ~mograsoJobSpectrum() {};
		virtual void job();
		virtual std::map<std::wstring, pointerWrapper<std::ofstream> >& get_fileMap();
		static std::vector<std::wstring> perJobFilenames();
		static std::vector<std::wstring> perVarSetFilenames();
	
	private:
		mograsoJobSpectrum();
		mograsoJobSpectrum(const mograsoJobSpectrum&);
};

template <class modeModel, class numT>
mograsoJobSpectrum<modeModel, numT>::mograsoJobSpectrum(const std::map<std::wstring, std::complex<double> >& varSet, const std::vector<layerDescription<numT> >& layers, const std::vector<section>& sections, const std::map<std::wstring, pointerWrapper<std::ofstream> >& fileMap, const std::wstring& lineStart, const std::map<std::wstring, std::wstring>& options)
{
	f_varSet = varSet;
	f_layers = layers;
	f_sections = sections;
	f_fileMap = fileMap;
	f_lineStart = lineStart;
	f_options = options;
}

template <class modeModel, class numT>
void mograsoJobSpectrum<modeModel, numT>::job()
{
	try
	{
		numT lambda = static_cast<numT>(NAN);
		numT scaling = static_cast<numT>(1e6);
		numT couplingIntervalMultiplier = static_cast<numT>(1e-3);
		
		if (f_fileMap.count(L"log") > 0)
		{
			pointerWrapper<std::ofstream> log = f_fileMap.find(L"log")->second;
			log.lock();
			*(log.getObjectPtr()) << PACKAGE_STRING << "\n\n";
			*(log.getObjectPtr()) << "The following options were chosen:\n";
			for (std::map<std::wstring, std::wstring>::iterator it = f_options.begin(); it != f_options.end(); it++)
				*(log.getObjectPtr()) << wstos(it->first) << ": " << wstos(it->second) << "\n";
			*(log.getObjectPtr()) << "\n";
			log.unlock();			
		}

		if (f_varSet.count(L"wavelength") > 0)
			lambda = std::abs(f_varSet[L"wavelength"]);
		else
			return;
		
		if (f_fileMap.count(L"log") > 0)
		{
			pointerWrapper<std::ofstream> log = f_fileMap.find(L"log")->second;
			log.lock();
			*(log.getObjectPtr()) << "Wavelength: " << doubleValue(lambda) << "\n\n";
			log.unlock();			
		}		
		
		if (f_varSet.count(L"couplingIntervalMultiplier") > 0)
			couplingIntervalMultiplier = std::abs(f_varSet[L"couplingIntervalMultiplier"]);
		
		if (f_fileMap.count(L"log") > 0)
		{
			pointerWrapper<std::ofstream> log = f_fileMap.find(L"log")->second;
			log.lock();
			*(log.getObjectPtr()) << "Layers:\n";
			for (typename std::vector<layerDescription<numT> >::iterator it = f_layers.begin(); it != f_layers.end(); it++)
				*(log.getObjectPtr()) << "r: " << doubleValue(it->r) << ", e: " << doubleValue(it->e) << ", u: " << doubleValue(it->u) << "\n";
			*(log.getObjectPtr()) << "\n";
			log.unlock();
		}
		
		modeSolver<modeModel, numT> solver(f_layers, lambda, scaling, static_cast<numT>(1e-12), static_cast<numT>(1e-5), &f_fileMap);
		mode<modeModel, numT> coreMode = solver.determine_core_mode();
		
		numT minPeriod = static_cast<numT>(NAN);
		numT maxPeriod = static_cast<numT>(NAN);
		numT minDC = static_cast<numT>(NAN);
		numT maxDC = static_cast<numT>(NAN);
		numT minP = static_cast<numT>(NAN);
		numT maxP = static_cast<numT>(NAN);
		for (std::vector<section>::iterator it = f_sections.begin(); it != f_sections.end(); it++)
		{
			if (std::isnan(minPeriod))
				minPeriod = static_cast<numT>(it->get_min_period());
			else
				minPeriod = std::min(minPeriod, static_cast<numT>(it->get_min_period()));
				
			if (std::isnan(maxPeriod))
				maxPeriod = it->get_max_period();
			else
				maxPeriod = std::max(maxPeriod, static_cast<numT>(it->get_max_period()));

			if (std::isnan(maxDC))
				maxDC = static_cast<numT>(it->get_dc_change());
			else
				maxDC = std::max(maxDC, static_cast<numT>(it->get_dc_change()));

			if (std::isnan(minDC))
				minDC = static_cast<numT>(it->get_dc_change());
			else
				minDC = std::min(minDC, static_cast<numT>(it->get_dc_change()));

			if (std::isnan(maxP))
				maxP = static_cast<numT>(it->get_p_change());
			else
				maxP = std::max(maxP, static_cast<numT>(it->get_p_change()));

			if (std::isnan(minP))
				minP = static_cast<numT>(it->get_p_change());
			else
				minP = std::min(minP, static_cast<numT>(it->get_p_change()));
		}
		
		numT k0Scaled = static_cast<numT>(2.0 * M_PI) / (lambda * scaling);
		coreMode.set_scaling(scaling);
		coreMode.set_normalisation(static_cast<numT>(1.0));
		std::complex<numT> couplingIntegral = coupling_integral_radial(coreMode, coreMode, static_cast<numT>(0.0), f_layers[0].r);
		numT couplingFactorMax = maxDC * std::real((k0Scaled * static_cast<std::complex<numT> >(std::sqrt(kEPSILON_0 / kMU_0)) / (scaling * scaling)) / (static_cast<numT>(2.0 * 1.0)) * f_layers[0].n * maxP * static_cast<numT>(2.0 * M_PI) * couplingIntegral);
		numT couplingFactorMin = minDC * std::real((k0Scaled * static_cast<std::complex<numT> >(std::sqrt(kEPSILON_0 / kMU_0)) / (scaling * scaling)) / (static_cast<numT>(2.0 * 1.0)) * f_layers[0].n * minP * static_cast<numT>(2.0 * M_PI) * couplingIntegral);
		numT intervalMean1 = std::real(coreMode.get_nEff() - (lambda * scaling) / (maxPeriod * scaling));
		numT intervalMean2 = std::real(coreMode.get_nEff() - (lambda * scaling) / (minPeriod * scaling));
		numT realStart1 = intervalMean1 + (couplingIntervalMultiplier * static_cast<numT>(2.0) + couplingFactorMax) / k0Scaled;
		numT realStart2 = intervalMean2 + (couplingIntervalMultiplier * static_cast<numT>(2.0) + couplingFactorMax) / k0Scaled;
		numT realEnd1 = intervalMean1 - (couplingIntervalMultiplier / static_cast<numT>(2.0) + couplingFactorMin) / k0Scaled;
		numT realEnd2 = intervalMean1 - (couplingIntervalMultiplier / static_cast<numT>(2.0) + couplingFactorMin) / k0Scaled;
			
		interval<numT> imag_interval = {static_cast<numT>(2e-5), static_cast<numT>(-1e-5)};
		for (typename std::vector<layerDescription<numT> >::iterator it = f_layers.begin(); it != f_layers.end(); it++)
		{
			imag_interval.start_T = std::max(imag_interval.start_T, std::imag(it->n));
			imag_interval.end_T = std::min(imag_interval.end_T, std::imag(it->n));
		}
		if (imag_interval.start_T > static_cast<numT>(1e-2))
			imag_interval.start_T = static_cast<numT>(1e-2);
		if (imag_interval.end_T > static_cast<numT>(-5e-3))
			imag_interval.end_T = static_cast<numT>(-5e-3);			
		if (std::abs(imag_interval.start_T) == std::abs(imag_interval.start_T))
			imag_interval.start_T *= static_cast<numT>(1.5);

		interval<std::complex<numT> > n_interval;
		n_interval.start_T = std::complex<numT>(std::max(realStart1, realStart2), std::max(imag_interval.start_T, imag_interval.end_T));
		n_interval.end_T = std::complex<numT>(std::min(realEnd1, realEnd2), std::min(imag_interval.start_T, imag_interval.end_T));

		if (f_fileMap.count(L"log") > 0)
		{
			pointerWrapper<std::ofstream> log = f_fileMap.find(L"log")->second;
			log.lock();
			*(log.getObjectPtr()) << "Mode interval: " << doubleValue(n_interval.start_T) << "-" << doubleValue(n_interval.end_T) << "\n\n";
			log.unlock();
		}
	
		std::vector<mode<modeModel, numT> > modes = solver.determine_modes_in_interval(n_interval);
		
		modes.insert(modes.begin(), coreMode);

		if (f_fileMap.count(L"modes") > 0)
		{
			pointerWrapper<std::ofstream> log = f_fileMap.find(L"modes")->second;
			log.lock();
			log.getObjectPtr()->precision(16);
			*(log.getObjectPtr()) << wstos(f_lineStart);
			for (int counter = 0; counter < modes.size(); counter++)
				*(log.getObjectPtr()) << doubleValue(std::real(modes[counter].get_nEff())) << "\t" << doubleValue(std::imag(modes[counter].get_nEff())) << "\t";
			*(log.getObjectPtr()) << "\n";
			log.unlock();
		}
		if (f_fileMap.count(L"log") > 0)
		{
			pointerWrapper<std::ofstream> log = f_fileMap.find(L"log")->second;
			log.lock();
			log.getObjectPtr()->precision(16);
			*(log.getObjectPtr()) << "Modes found:\n";
			for (int counter = 0; counter < modes.size(); counter++)
				*(log.getObjectPtr()) << doubleValue(std::real(modes[counter].get_nEff())) << "\t" << doubleValue(std::imag(modes[counter].get_nEff())) << "\t";
			*(log.getObjectPtr()) << "\n";
			log.unlock();
		}
		
		// Initiate mode amplitudes
		matrix<std::complex<numT> > amplitudes(modes.size());
		amplitudes(0) = 1.0;
		
		for (std::vector<section>::iterator it = f_sections.begin(); it != f_sections.end(); it++)
		{
			if ((f_options.count(L"gratingModel") > 0) && (f_options.find(L"gratingModel")->second == L"TM"))
			{
				gratingModelTM<modeModel, numT> gSolver(modes, *it, lambda, scaling, static_cast<numT>(1.0));
				amplitudes = gSolver.amplitudes(amplitudes);
			}
			else
			{
				gratingModelCMT<modeModel, numT> gSolver(modes, *it, lambda, scaling, static_cast<numT>(1.0));
				amplitudes = gSolver.amplitudes(amplitudes);
			}
		}
		
		numT transmittance = static_cast<numT>(10.0) * log10(std::abs(amplitudes(0) * amplitudes(0)));

		if (f_fileMap.count(L"transmittance") > 0)
		{
			pointerWrapper<std::ofstream> log = f_fileMap.find(L"transmittance")->second;
			log.lock();
			log.getObjectPtr()->precision(16);
			*(log.getObjectPtr()) << wstos(f_lineStart) << doubleValue(transmittance) << "\n";
			log.unlock();
		}
	}
	catch (std::runtime_error e)
	{
		if (f_fileMap.count(L"log") > 0)
		{
			pointerWrapper<std::ofstream> log = f_fileMap.find(L"log")->second;
			log.lock();
			*(log.getObjectPtr()) << e.what() << "\n";
			log.unlock();
		}		
		if (f_fileMap.count(L"transmittance") > 0)
		{
			pointerWrapper<std::ofstream> log = f_fileMap.find(L"transmittance")->second;
			log.lock();
			log.getObjectPtr()->precision(16);
			*(log.getObjectPtr()) << wstos(f_lineStart) << "NaN\n";
			log.unlock();
		}
	}
	catch (...)
	{
		if (f_fileMap.count(L"log") > 0)
		{
			pointerWrapper<std::ofstream> log = f_fileMap.find(L"log")->second;
			log.lock();
			*(log.getObjectPtr()) << "Error\n";
			log.unlock();
		}		
		if (f_fileMap.count(L"transmittance") > 0)
		{
			pointerWrapper<std::ofstream> log = f_fileMap.find(L"transmittance")->second;
			log.lock();
			log.getObjectPtr()->precision(16);
			*(log.getObjectPtr()) << wstos(f_lineStart) << "NaN\n";
			log.unlock();
		}
	}	
}

template <class modeModel, class numT>
std::map<std::wstring, pointerWrapper<std::ofstream> >& mograsoJobSpectrum<modeModel, numT>::get_fileMap()
{
	return f_fileMap;
}

template <class modeModel, class numT>
std::vector<std::wstring> mograsoJobSpectrum<modeModel, numT>::perJobFilenames()
{
	std::vector<std::wstring> out;
	out.push_back(L"modes");
	out.push_back(L"transmittance");
	return out;
}

template <class modeModel, class numT>
std::vector<std::wstring> mograsoJobSpectrum<modeModel, numT>::perVarSetFilenames()
{
	std::vector<std::wstring> out;
	out.push_back(L"log");
	return out;
}

#endif

