/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#include "variableStorage.h"

#pragma mark --- variable ---

variable::variable(const std::wstring& name, const std::complex<double>& start, const std::complex<double>& end, const unsigned long& numberOfSteps, const std::wstring& prefix, const std::wstring& suffix)
{
	f_name = name;
	f_start = start;
	f_end = end;
	f_numberOfSteps = numberOfSteps;
	f_prefix = prefix;
	f_suffix = suffix;
	
	int precision = 1;
	while ((precision < 16) && (complexDoubleToWString(f_start, precision) == complexDoubleToWString(f_start + (f_end - f_start) / (double)f_numberOfSteps, precision)))
		precision++;

	// Take care of rounding and assign
	f_requiredPrecision = ++precision;
}

variable::variable(const variable& obj)
{
	f_name = obj.f_name;
	f_start = obj.f_start;
	f_end = obj.f_end;
	f_numberOfSteps = obj.f_numberOfSteps;
	f_prefix = obj.f_prefix;
	f_suffix = obj.f_suffix;
	f_requiredPrecision = obj.f_requiredPrecision;
}

variable::~variable()
{
}

std::wstring variable::get_name() const
{
	return f_name;
}

unsigned long variable::get_numberOfValues() const
{
	return f_numberOfSteps + 1;
}

std::complex<double> variable::get_value(const unsigned long& number) const
{
	if (f_numberOfSteps == 0)
		return f_start;
	else
		return (double)number / (double)f_numberOfSteps * (f_end - f_start) + f_start;
}

std::wstring variable::get_prefix() const
{
	return f_prefix;
}

std::wstring variable::get_suffix() const
{
	return f_suffix;
}

int variable::get_requiredPrecision() const
{
	return f_requiredPrecision;
}

#pragma mark --- variableStorage ---
variableStorage::variableStorage()
{
	f_current = 0;
}

variableStorage::variableStorage(const variableStorage& obj)
{
	f_current = 0;
}

variableStorage::~variableStorage()
{
}

void variableStorage::add_variable(const std::wstring& name, const std::complex<double>& start, const std::complex<double>& end, const unsigned long& numberOfSteps, const std::wstring& prefix, const std::wstring& suffix)
{
	add_variable(variable(name, start, end, numberOfSteps, prefix, suffix));
}

void variableStorage::add_variable(const variable& var)
{
	f_vars.push_back(var);
}

void variableStorage::init(const bool& randomise)
{
	unsigned long numberOfVarSets = get_numberOfVarSets();

	std::map<unsigned long, unsigned long> temp;
	for (unsigned long counter = 0; counter < numberOfVarSets; counter++)
		temp[counter] = counter;
		
	f_order = temp;
}

unsigned long variableStorage::get_numberOfVarSets() const throw (std::runtime_error)
{
	unsigned long numberOfVarSets = 0;
	
	for (std::vector<variable>::const_iterator it = f_vars.begin(); it != f_vars.end(); it++)
		if (it == f_vars.begin())
			numberOfVarSets = it->get_numberOfValues();
		else
			numberOfVarSets *= it->get_numberOfValues();
			
	return numberOfVarSets;
}

std::map<std::wstring, std::complex<double> > variableStorage::get_varSet(const unsigned long& number) const throw (std::runtime_error)
{
	if (!is_init())
		throw std::runtime_error("variableStorage::get_varSet: variableStorage is not initialised.");

	if (number > (f_order.size() - 1))
		throw std::runtime_error("variableStorage::get_varSet: Number beyond bounds.");

	std::map<std::wstring, std::complex<double> > varSet;
	unsigned long varNum = number;
	
	for (std::vector<variable>::const_iterator it = f_vars.begin(); it != f_vars.end(); it++)
	{
		unsigned long num = varNum % it->get_numberOfValues();
		varSet[it->get_name()] = it->get_value(num);
		varNum = (varNum - num) / it->get_numberOfValues();
	}
	
	return varSet;
}

std::map<std::wstring, std::complex<double> > variableStorage::get_nextVarSet() throw (std::runtime_error)
{
	return get_varSet(++f_current);
}

std::wstring variableStorage::get_filenameStringForVarSet(const unsigned long& number) const throw (std::runtime_error)
{
	std::map<std::wstring, std::complex<double> > varSet = get_varSet(number);
	std::wstringstream strStream;
	
	for (std::vector<variable>::const_iterator it = f_vars.begin(); it != f_vars.end(); it++)
		if ((it->get_numberOfValues() > 1) && (varSet.count(it->get_name()) > 0))
			strStream << L"_" << it->get_prefix() << complexDoubleToWString(varSet.find(it->get_name())->second, it->get_requiredPrecision()) << it->get_suffix();
			
	// Mask signs that might cause trouble (i.e. /\.:).
	std::wstring str(strStream.str());
	for (std::wstring::size_type pos = str.find(L"/"); pos != std::wstring::npos; pos = str.find(L"/"))
		str.replace(pos, 1, L"_");
	for (std::wstring::size_type pos = str.find(L"\\"); pos != std::wstring::npos; pos = str.find(L"\\"))
		str.replace(pos, 1, L"_");
	for (std::wstring::size_type pos = str.find(L"."); pos != std::wstring::npos; pos = str.find(L"."))
		str.replace(pos, 1, L"_");
	for (std::wstring::size_type pos = str.find(L":"); pos != std::wstring::npos; pos = str.find(L":"))
		str.replace(pos, 1, L"_");
	
	return str;
}

std::wstring variableStorage::get_filenameStringForCurrentVarSet() const throw (std::runtime_error)
{
	return get_filenameStringForVarSet(f_current);
}

std::wstring variableStorage::get_lineStartForVarSet(const unsigned long& number) const throw (std::runtime_error)
{
	std::map<std::wstring, std::complex<double> > varSet = get_varSet(number);
	std::wstringstream strStream;
	
	for (std::vector<variable>::const_iterator it = f_vars.begin(); it != f_vars.end(); it++)
		if ((it->get_numberOfValues() > 1) && (varSet.count(it->get_name()) > 0))
		{
			std::complex<double> temp(varSet.find(it->get_name())->second);
			strStream << complexDoubleToWString(std::real(temp), 16) << L"\t" << complexDoubleToWString(std::imag(temp), 16) << L"\t";
		}
			
	return strStream.str();
}

bool variableStorage::is_init() const
{
	if (f_current > (f_order.size() - 1))
		return false;
		
	return true;
}
