/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#ifndef _AD_H
#define _AD_H 1

#include <cmath>
#include <stdexcept>
#include <complex>
//#include "bessel_amos.h"

template <class T>
class AD
{
	private:
		T _fun_T;
		T _diff_T;
		
	public:
		AD(const T& fun = static_cast<T>(0.0), const T& diff = static_cast<T>(0.0));
		AD(const double& fun);
		T& fun(void);
		T& diff(void);
		AD<T> operator+(const AD<T>& op) const;
		void operator+=(const AD<T>& op);
		AD<T> operator-(const AD<T>& op) const;
		void operator-=(const AD<T>& op);
		AD<T> operator-(void) const;
		AD<T> operator*(const AD<T>& op) const;
		void operator*=(const AD<T>& op);
		AD<T> operator/(const AD<T>& op) const throw (std::runtime_error);
		void operator/=(const AD<T>& op) throw (std::runtime_error);
		AD<T> sin(void) const;
		AD<T> cos(void) const;
		AD<T> exp(void) const;
		AD<T> log(void) const throw (std::runtime_error);
		AD<T> sqrt(void) const;
		T abs(void) const;
#ifdef _HH_BESSEL_AMOS
		AD<T> besselj(const double& order);
		AD<T> bessely(const double& order);
		AD<T> besseli(const double& order);
		AD<T> besselk(const double& order);
		AD<T> besselh1(const double& order);
		AD<T> besselh2(const double& order);
#endif
};

inline double posDoubleRepresentation(AD<double> op) { return std::abs(op.fun()); };
inline double posDoubleRepresentation(AD<std::complex<double> > op) { return std::abs(op.fun()); };

template <class T>
AD<T>::AD(const T& fun, const T& diff)
{
	_fun_T = fun;
	_diff_T = diff;
}

template <class T>
AD<T>::AD(const double& fun)
{
	_fun_T = static_cast<T>(fun);
}

template <class T>
T& AD<T>::fun(void)
{
	return _fun_T;
}

template <class T>
T& AD<T>::diff(void)
{
	return _diff_T;
}

template <class T>
AD<T> AD<T>::operator+(const AD<T>& op) const
{
	AD<T> ret;
	ret._fun_T = _fun_T + op._fun_T;
	ret._diff_T = _diff_T + op._diff_T;
	return ret;
}

template <class T>
void AD<T>::operator+=(const AD<T>& op)
{
	_fun_T += op._fun_T;
	_diff_T += op._diff_T;
}

template <class T>
AD<T> AD<T>::operator-(const AD<T>& op) const
{
	AD<T> ret;
	ret._fun_T = _fun_T - op._fun_T;
	ret._diff_T = _diff_T - op._diff_T;
	return ret;
}

template <class T>
void AD<T>::operator-=(const AD<T>& op)
{
	_fun_T -= op._fun_T;
	_diff_T -= op._diff_T;
}

template <class T>
AD<T> AD<T>::operator-() const
{
	AD<T> ret;
	ret._fun_T = -_fun_T;
	ret._diff_T = -_diff_T;
	return ret;
}

template <class T>
AD<T> AD<T>::operator*(const AD<T>& op) const
{
	AD<T> ret;
	ret._fun_T = _fun_T * op._fun_T;
	ret._diff_T = _diff_T * op._fun_T + _fun_T * op._diff_T;
	
	return ret;
}

template <class T>
void AD<T>::operator*=(const AD<T>& op)
{
	T fun = _fun_T;
	T diff = _diff_T;
	
	_fun_T = fun * op._fun_T;
	_diff_T = diff * op._fun_T + fun * op._diff_T;
}

template <class T>
AD<T> AD<T>::operator/(const AD<T>& op) const throw (std::runtime_error)
{
 	if (std::abs(op._fun_T) == static_cast<T>(0.0))
 		throw std::runtime_error("AD<T>::operator/(const AD<T>& op): Division through zero.");

	AD<T> ret;
	ret._fun_T = _fun_T / op._fun_T;
	ret._diff_T = (_diff_T * op._fun_T - _fun_T * op._diff_T) / (op._fun_T * op._fun_T);
	
	return ret;
}

template <class T>
void AD<T>::operator/=(const AD<T>& op) throw (std::runtime_error)
{
 	if (std::abs(op._fun_T) == static_cast<T>(0.0))
 		throw std::runtime_error("AD<T>::operator/(const AD<T>& op): Division through zero.");

	T fun = _fun_T;
	T diff = _diff_T;

	_fun_T = fun / op._fun_T;
	_diff_T = (diff * op._fun_T - fun * op._diff_T) / (op._fun_T * op._fun_T);
}

template <class T>
AD<T> AD<T>::sin() const
{
	AD<T> ret;
	ret._fun_T = std::sin(_fun_T);
	ret._diff_T = _diff_T * std::cos(_fun_T);
	
	return ret;
}

template <class T>
AD<T> AD<T>::cos() const
{
	AD<T> ret;
	ret._fun_T = std::cos(_fun_T);
	ret._diff_T = -_diff_T * std::sin(_fun_T);
	
	return ret;
}

template <class T>
AD<T> AD<T>::exp() const
{
	AD<T> ret;
	ret._fun_T = std::exp(_fun_T);
	ret._diff_T = _diff_T * std::exp(_fun_T);
	
	return ret;	
}

template <class T>
AD<T> AD<T>::log() const throw (std::runtime_error)
{
	if (_fun_T == static_cast<T>(0.0))
		throw std::runtime_error("AD<T>::log(): Division throught zero and log of zero.");

	AD<T> ret;
	ret._fun_T = std::log(_fun_T);
	ret._diff_T = _diff_T / _fun_T;
	
	return ret;	
}

template <class T>
AD<T> AD<T>::sqrt() const
{
	AD<T> ret;
	ret._fun_T = std::sqrt(_fun_T);
	ret._diff_T = _diff_T / (static_cast<T>(2.0) * std::sqrt(_fun_T));
	
	return ret;	
}

template <class T>
T AD<T>::abs() const
{
	return std::abs(_fun_T);	
}

#ifdef _HH_BESSEL_AMOS
template <class T>
AD<T> AD<T>::besselj(const double& order)
{
	AD<T> ret;
	T J = ::besselj(order, _fun_T);
	ret._fun_T = J;
	ret._diff_T = _diff_T * (order / _fun_T * J - ::besselj(order + 1.0, _fun_T));
	
	return ret;
}

template <class T>
AD<T> AD<T>::bessely(const double& order)
{
	AD<T> ret;
	T Y = ::bessely(order, _fun_T);
	ret._fun_T = Y;
	ret._diff_T = _diff_T * (order / _fun_T * Y - ::bessely(order + 1.0, _fun_T));
	
	return ret;
}

template <class T>
AD<T> AD<T>::besseli(const double& order)
{
	AD<T> ret;
	T I = ::besseli(order, _fun_T);
	ret._fun_T = I;
	ret._diff_T = _diff_T * (order / _fun_T * I + ::besseli(order + 1.0, _fun_T));
	
	return ret;
}

template <class T>
AD<T> AD<T>::besselk(const double& order)
{
	AD<T> ret;
	T K = ::besselk(order, _fun_T);
	ret._fun_T = K;
	ret._diff_T = _diff_T * (order / _fun_T * K - ::besselk(order + 1.0, _fun_T));
	
	return ret;
}

template <class T>
AD<T> AD<T>::besselh1(const double& order)
{
	AD<T> ret;
	T H1 = ::besselh1(order, _fun_T);
	ret._fun_T = H1;
	ret._diff_T = _diff_T * (order / _fun_T * H1 - ::besselh1(order + 1.0, _fun_T));
	
	return ret;
}

template <class T>
AD<T> AD<T>::besselh2(const double& order)
{
	AD<T> ret;
	T H2 = ::besselh2(order, _fun_T);
	ret._fun_T = H2;
	ret._diff_T = _diff_T * (order / _fun_T * H2 - ::besselh2(order + 1.0, _fun_T));
	
	return ret;
}
#endif

namespace std
{
template <class T>
AD<T> sin(AD<T> op) { return op.sin(); }

template <class T>
AD<T> cos(AD<T> op) { return op.cos(); }

template <class T>
AD<T> exp(AD<T> op) { return op.exp(); }

template <class T>
AD<T> log(AD<T> op) { return op.log(); }

template <class T>
AD<T> sqrt(AD<T> op) { return op.sqrt(); }

template <class T>
T abs(AD<T> op) { return op.abs(); }
}

#ifdef _HH_BESSEL_AMOS
template <class T>
AD<T> besselj(const double& order, AD<T> op) { return op.besselj(order); }

template <class T>
AD<T> bessely(const double& order, AD<T> op) { return op.bessely(order); }

template <class T>
AD<T> besseli(const double& order, AD<T> op) { return op.besseli(order); }

template <class T>
AD<T> besselk(const double& order, AD<T> op) { return op.besselk(order); }

template <class T>
AD<T> besselh1(const double& order, AD<T> op) { return op.besselh1(order); }

template <class T>
AD<T> besselh2(const double& order, AD<T> op) { return op.besselh2(order); }
#endif
#endif
