/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#include "jobServer.h"

unsigned long jobServer::f_current = 0;
unsigned long jobServer::f_stepsCompleted = 0;
int jobServer::f_threadCount = 0;
jobFactoryInterface* jobServer::f_jobFactory = 0;
time_t jobServer::f_startTime = 0;
pthread_mutex_t jobServer::f_timing_mutex;
pthread_mutex_t jobServer::f_jobFactory_mutex;

void jobServer::set_jobFactory(jobFactoryInterface* jobFactory)
{
	f_jobFactory = jobFactory;
}

void jobServer::do_jobs(const int& numberOfThreads)
{
	f_threadCount = numberOfThreads;
	pthread_mutex_init(&f_timing_mutex, NULL);
	pthread_mutex_init(&f_jobFactory_mutex, NULL);
	jobClient clients[f_threadCount];
	pthread_attr_t attr;		
	
	// Init the counter
	f_current = 0;
	
	// Init the timer
	f_startTime = time(NULL);
	
	// Init the completed jobs
	f_stepsCompleted = 0;
	
	// Init the thread attribute which is necessary to make the threads joinable.
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	
	// Detach the threads
	for (int threadCounter_int = 0; threadCounter_int < f_threadCount; threadCounter_int++)
	{
		clients[threadCounter_int].set_id(threadCounter_int);
		int result = pthread_create((clients[threadCounter_int].get_threadPtr()), &attr, clients[threadCounter_int].process, &clients[threadCounter_int]);
		if (result != 0)
			exit(-1);
	}
	
	// Join all the threads
	int status;
	for(int threadCounter_int = 0; threadCounter_int < f_threadCount; threadCounter_int++)
		pthread_join(*(clients[threadCounter_int].get_threadPtr()), (void **)&status);

	pthread_attr_destroy(&attr);
	pthread_mutex_destroy(&f_jobFactory_mutex);
	pthread_mutex_destroy(&f_timing_mutex);
}

pthread_t* jobClient::get_threadPtr()
{
	return &_thread;
}

void* jobClient::process(void* param)
{
	while (1)
	{
		jobInterface* job = 0;
		
		if (f_current < f_jobFactory->get_numberOfJobs())
		{
			pthread_mutex_lock(&f_jobFactory_mutex);
			job = f_jobFactory->checkout_job(f_current++);
			pthread_mutex_unlock(&f_jobFactory_mutex);
		}
		else
		{
			return 0;
		}
		
		if (job != 0)
		{
			try
			{
				job->job();
			}
			catch (...)
			{
				f_jobFactory->checkin_job(job);
				throw;
			}
			
			pthread_mutex_lock(&f_timing_mutex);
			f_stepsCompleted++;
			if ((f_stepsCompleted == f_jobFactory->get_numberOfJobs()) || (f_stepsCompleted % f_threadCount == 0))
			{
				double percentage = (double)f_stepsCompleted / (double)f_jobFactory->get_numberOfJobs();
				char buffer[33];
				sprintf(buffer, "%2.1f", percentage * 100.0);
				std::cout << std::string(buffer) << "% completed";
				if (percentage >= 0.3)
					std::cout << ", time remaining: " << seconds_to_string(difftime(time(NULL), f_startTime) * (1.0 / percentage - 1.0));
				std::cout << "\n";
				
				if (f_stepsCompleted == f_jobFactory->get_numberOfJobs())
					std::cout << "Total time: " << seconds_to_string(difftime(time(NULL), f_startTime)) << "\n";
			}
			pthread_mutex_unlock(&f_timing_mutex);
		}
		f_jobFactory->checkin_job(job);
	}	
}

void jobClient::set_id(int id)
{
	_id = id;
}

const int& jobClient::get_id() const
{
	return _id;
}
