/*
 * MoGraSo: A programme for the calculation of the response of fibre grating
 * elements using modal methods.
 *
 * MoGraSo Copyright (C) 2007 Hannes Hochreiner <hannes@hochreiner.net>
 *
 * This file is part of MoGraSo.
 *
 * MoGraSo is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * MoGraSo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * MoGraSo.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#include "matrix_template.h"

#ifdef HAVE_LAPACK
void indexArray_from_permutationArray(const unsigned int& dimension, __CLPK_integer* permutationArray, unsigned int* indexArray)
{
	for (unsigned int count = 0; count < dimension; count++)
		indexArray[count] = count;

	for (unsigned int count = 0; count < dimension; count++)
	{
		if (count != (permutationArray[count] - 1))
		{
			indexArray[count] = (unsigned int)(permutationArray[count]) - 1;
			indexArray[permutationArray[count] - 1] = count;
		}
	}
}

std::complex<double> operator+ (const std::complex<double> op1, const matrixNumber& op2)
{
	return op2.operator+(op1);
}

std::complex<double> operator- (const std::complex<double> op1, const matrixNumber& op2)
{
	return op2.operator-(op1);
}

std::complex<double> operator* (const std::complex<double> op1, const matrixNumber& op2)
{
	return op2.operator*(op1);
}

std::complex<double> operator/ (const std::complex<double> op1, const matrixNumber& op2)
{
	return op2.operator/(op1);
}

std::complex<double> operator+ (const double op1, const matrixNumber& op2)
{
	return op2.operator+(op1);
}

std::complex<double> operator- (const double op1, const matrixNumber& op2)
{
	return op2.operator-(op1);
}

std::complex<double> operator* (const double op1, const matrixNumber& op2)
{
	return op2.operator*(op1);
}

std::complex<double> operator/ (const double op1, const matrixNumber& op2)
{
	return op2.operator/(op1);
}

std::ostream& operator<< (std::ostream& os, const matrixNumber& number)
{
	return number.output(os);
}

namespace std
{
	complex<double> conj(const matrixNumber& number)
	{
		complex<double> temp(number.conj());
		return temp;
	}
	
	double abs(const matrixNumber& number)
	{
		double temp(number.abs());
		return temp;
	}
	
	double real(const matrixNumber& number)
	{
		double temp(number.real());
		return temp;
	}
	
	double imag(const matrixNumber& number)
	{
		double temp(number.imag());
		return temp;
	}
	
	void operator*= (complex<double>& op1, const matrixNumber& number)
	{
		complex<double> temp(number.real(), number.imag());
		op1 *= temp;
	}
}

matrix<std::complex<double> >::matrix(void)
{
	this->_data = NULL;
	this->_ipiv = NULL;
	this->_rowPermutation_uintPtr = NULL;
	this->_rowCount_uint = 0;
	this->_columnCount_uint = 0;
}

matrix<std::complex<double> >::matrix(const unsigned int& rows, const unsigned int& columns, const std::complex<double>& value, const fillPattern_enum& fillPattern)
{
	this->allocate(rows, columns);
	this->fill(value, fillPattern);
}

matrix<std::complex<double> >::matrix(const matrix<std::complex<double> >& op) throw (std::runtime_error)
{
	this->allocate(op._rowCount_uint, op._columnCount_uint);

	for (unsigned int rowCounter_uint = 0; rowCounter_uint < this->_rowCount_uint; rowCounter_uint++)
		for (unsigned int columnCounter_uint = 0; columnCounter_uint < this->_columnCount_uint; columnCounter_uint++)
			(*this)(rowCounter_uint, columnCounter_uint) = op(rowCounter_uint, columnCounter_uint);

}

matrix<std::complex<double> >::~matrix(void)
{
	this->deallocate();
}

void matrix<std::complex<double> >::allocate(const unsigned int& rows, const unsigned int& columns) throw (std::runtime_error)
{
	if ((rows > 0) && (columns > 0))
	{
		this->_rowCount_uint = rows;
		this->_columnCount_uint = columns;
		this->_data = new __CLPK_doublecomplex[this->_rowCount_uint * this->_columnCount_uint];
		this->_ipiv = new __CLPK_integer[this->_rowCount_uint];
		this->_rowPermutation_uintPtr = new unsigned int[this->_rowCount_uint];

		for (unsigned int rowCounter_uint = 0; rowCounter_uint < this->_rowCount_uint; rowCounter_uint++)
		{
			this->_ipiv[rowCounter_uint] = rowCounter_uint;
			this->_rowPermutation_uintPtr[rowCounter_uint] = rowCounter_uint;
		}
	}
	else
	{
		this->_rowCount_uint = 0;
		this->_columnCount_uint = 0;
		this->_data = NULL;
		this->_ipiv = NULL;
		throw std::runtime_error("matrix::matrix allocation failed");
	}
}

void matrix<std::complex<double> >::deallocate(void)
{
	if ((this->_rowCount_uint > 0) && (this->_data != NULL))
	{
		delete[] this->_data;
		this->_data = NULL;
		this->_rowCount_uint = 0;
		this->_columnCount_uint = 0;
		delete[] this->_rowPermutation_uintPtr;
		this->_rowPermutation_uintPtr = NULL;
		delete[] this->_ipiv;
		this->_ipiv = NULL;
	}
}

void matrix<std::complex<double> >::operator= (const matrix<std::complex<double> >& op)
{
	this->deallocate();
	this->allocate(op._rowCount_uint, op._columnCount_uint);

	for (unsigned int rowCounter_uint = 0; rowCounter_uint < this->_rowCount_uint; rowCounter_uint++)
		for (unsigned int columnCounter_uint = 0; columnCounter_uint < this->_columnCount_uint; columnCounter_uint++)
			(*this)(rowCounter_uint, columnCounter_uint) = op(rowCounter_uint, columnCounter_uint);
}

matrixNumber matrix<std::complex<double> >::operator() (const unsigned int& row, const unsigned int& column) const throw (std::runtime_error)
{
	if ((row < _rowCount_uint) && (column < _columnCount_uint))
		return matrixNumber(&(_data[_columnCount_uint * row + column]));
	else
		throw std::runtime_error("matrix operator() indices out of range");
}

matrix<std::complex<double> > matrix<std::complex<double> >::operator* (const matrix<std::complex<double> >& op1) throw (std::runtime_error)
{
	if (this->_columnCount_uint == op1._rowCount_uint)
	{
		matrix<std::complex<double> > temp(this->_rowCount_uint, op1._columnCount_uint);
		for (unsigned int lineCounter_uint = 0; lineCounter_uint < this->_rowCount_uint; lineCounter_uint++)
			for (unsigned int columnCounter_uint = 0; columnCounter_uint < op1._columnCount_uint; columnCounter_uint++)
				for (unsigned int innerCounter_uint = 0; innerCounter_uint < this->_columnCount_uint; innerCounter_uint++)
					temp(lineCounter_uint, columnCounter_uint) += (*this)(lineCounter_uint, innerCounter_uint) * op1(innerCounter_uint, columnCounter_uint);

		return temp;
	}
	else
	{
		throw std::runtime_error("matrix operator* inner dimensions do not agree");
	}
}

matrix<std::complex<double> > matrix<std::complex<double> >::operator* (const std::complex<double>& op1)
{
	matrix<std::complex<double> > temp(*this);
	for (unsigned int lineCounter_uint = 0; lineCounter_uint < this->_rowCount_uint; lineCounter_uint++)
		for (unsigned int columnCounter_uint = 0; columnCounter_uint < this->_columnCount_uint; columnCounter_uint++)
			temp(lineCounter_uint, columnCounter_uint) *= op1;

	return temp;	
}

matrix<std::complex<double> > matrix<std::complex<double> >::operator+ (const std::complex<double>& op1)
{
	matrix<std::complex<double> > temp(*this);
	for (unsigned int lineCounter_uint = 0; lineCounter_uint < this->_rowCount_uint; lineCounter_uint++)
		for (unsigned int columnCounter_uint = 0; columnCounter_uint < this->_columnCount_uint; columnCounter_uint++)
			temp(lineCounter_uint, columnCounter_uint) += op1;

	return temp;	
}

unsigned int matrix<std::complex<double> >::rows(void) const
{
	return this->_rowCount_uint;
}

unsigned int matrix<std::complex<double> >::columns(void) const
{
	return this->_columnCount_uint;
}

void matrix<std::complex<double> >::fill(const std::complex<double>& value, const fillPattern_enum& fillPattern)
{
	switch (fillPattern)
	{
		case all_fillPattern:
			for (unsigned int rowCounter_uint = 0; rowCounter_uint < this->_rowCount_uint; rowCounter_uint++)
			{
				for (unsigned int columnCounter_uint = 0; columnCounter_uint < this->_columnCount_uint; columnCounter_uint++)
				{
					(*this)(rowCounter_uint, columnCounter_uint) = value;
				}
			}
			break;
		case diagonal_fillPattern:
			for (unsigned int rowCounter_uint = 0; rowCounter_uint < this->_rowCount_uint; rowCounter_uint++)
			{
				for (unsigned int columnCounter_uint = 0; columnCounter_uint < this->_columnCount_uint; columnCounter_uint++)
				{
					if (rowCounter_uint == columnCounter_uint)
					{
						(*this)(rowCounter_uint, columnCounter_uint) = value;
					}
					else
					{
						(*this)(rowCounter_uint, columnCounter_uint) = 0.0;
					}
				}
			}
			break;
	}
}

matrix<std::complex<double> > matrix<std::complex<double> >::transpose(void) const
{
	matrix<std::complex<double> > output_matrix(this->_columnCount_uint, this->_rowCount_uint);
	
	for (unsigned int rowCounter_uint = 0; rowCounter_uint < output_matrix._rowCount_uint; rowCounter_uint++)
		for (unsigned int columnCounter_uint = 0; columnCounter_uint < output_matrix._columnCount_uint; columnCounter_uint++)
			output_matrix(rowCounter_uint, columnCounter_uint) = (*this)(columnCounter_uint, rowCounter_uint);
	
	return output_matrix;
}

matrix<std::complex<double> > matrix<std::complex<double> >::ludecomp(void) throw (std::runtime_error)
{
	__CLPK_integer m = (int)_rowCount_uint;
	__CLPK_integer n = (int)_columnCount_uint;
	__CLPK_integer info, result;
	matrix<std::complex<double> > temp(*this);
	
	result = zgetrf_(&m, &n, temp._data, &m, temp._ipiv, &info);
	
 	if (info < 0)
 		throw std::runtime_error("matrix<std::complex<double> > matrix<std::complex<double> >::ludecomp: One argument had an illegal value.");
 	else if (info > 0)
 		throw std::runtime_error("matrix<std::complex<double> > matrix<std::complex<double> >::ludecomp: Matrix is singular.");
	
	indexArray_from_permutationArray(_rowCount_uint, temp._ipiv, temp._rowPermutation_uintPtr);

	return temp;
}

matrix<std::complex<double> > matrix<std::complex<double> >::fbsubs(const matrix<std::complex<double> >& b) throw (std::runtime_error)
{
	__CLPK_integer info, result;
	__CLPK_integer m = (int)_rowCount_uint;
	__CLPK_integer n = (int)_columnCount_uint;
	matrix<std::complex<double> > temp(b);
	__CLPK_integer nrhs = (int)temp.columns();
	__CLPK_integer ldb = (int)temp.rows();
	char trans = 'N';
	
	result = zgetrs_(&trans, &n, &nrhs, _data, &m, _ipiv, temp._data, &ldb, &info);

 	if (info < 0)
 		throw std::runtime_error("matrix<std::complex<double> > matrix<std::complex<double> >::fbsubs: One argument had an illegal value.");
	
	return temp;
}

matrix<std::complex<double> > matrix<std::complex<double> >::inv(void)
{
	__CLPK_integer info, result, lwork;
	matrix<std::complex<double> > temp_matrix(this->ludecomp());
	__CLPK_integer n = (int)temp_matrix._columnCount_uint;
	__CLPK_integer lda = (int)temp_matrix.rows();
	__CLPK_doublecomplex workQuery;
	
	lwork = -1;
	result = zgetri_(&n, temp_matrix._data, &lda, temp_matrix._ipiv, &workQuery, &lwork, &info);
	lwork = (int)workQuery.r;
	__CLPK_doublecomplex work[lwork];
	result = zgetri_(&n, temp_matrix._data, &lda, temp_matrix._ipiv, work, &lwork, &info);

 	if (info < 0)
 		throw std::runtime_error("matrix<std::complex<double> > matrix<std::complex<double> >::inv: One argument had an illegal value.");
	
	return temp_matrix;
}

std::complex<double> matrix<std::complex<double> >::det(void)
{
	matrix<std::complex<double> > LU_matrix(this->ludecomp());
	std::complex<double> determinant;

	unsigned int permutation_sign = 0;
	
	for (unsigned int counter = 0; counter < LU_matrix.rows(); counter++)
	{
		if (counter == 0)
		{
			determinant = LU_matrix(counter, counter);
		} else {
			determinant *= LU_matrix(counter, counter);
		}

		if (counter != (unsigned int)(LU_matrix._ipiv[counter] - 1))
			permutation_sign++;
	}


	if ((permutation_sign) & 1) {
		return determinant * -1.0;
	} else {
		return determinant;
	}
}

matrix<std::complex<double> > matrix<std::complex<double> >::abs(void)
{
	matrix<std::complex<double> > output_matrix(*this);

	for (unsigned int rowCounter_uint = 0; rowCounter_uint < this->_rowCount_uint; rowCounter_uint++)
		for (unsigned int columnCounter_uint = 0; columnCounter_uint < this->_columnCount_uint; columnCounter_uint++)
			output_matrix(rowCounter_uint, columnCounter_uint) = std::abs(output_matrix(rowCounter_uint, columnCounter_uint));

	return output_matrix;
}

std::ostream& matrix<std::complex<double> >::output(std::ostream& os) const
{
	if ((this->_rowCount_uint > 0) && (this->_data != NULL))
		for (unsigned int rowCounter_uint = 0; rowCounter_uint < this->_rowCount_uint; rowCounter_uint++)
		{
			for (unsigned int columnCounter_uint = 0; columnCounter_uint < this->_columnCount_uint; columnCounter_uint++)
				os << (*this)(rowCounter_uint, columnCounter_uint) << "\t";
			os << "\n";
		}

	return os;
}

matrixNumber::matrixNumber(__CLPK_doublecomplex* const number)
{
	_number = number;
}

matrixNumber::operator std::complex<double>()
{ 
	std::complex<double> temp(_number->r, _number->i);
	return temp;
}

void matrixNumber::operator= (const matrixNumber& number)
{
	_number->r = number._number->r;
	_number->i = number._number->i;
}

void matrixNumber::operator= (const std::complex<double>& number)
{
	_number->r = std::real(number);
	_number->i = std::imag(number);
}

void matrixNumber::operator+= (const matrixNumber& number)
{
	_number->r += number._number->r;
	_number->i += number._number->i;
}

void matrixNumber::operator-= (const matrixNumber& number)
{
	_number->r -= number._number->r;
	_number->i -= number._number->i;
}

void matrixNumber::operator*= (const matrixNumber& number)
{
	std::complex<double> op1(_number->r, _number->i);
	std::complex<double> op2(number._number->r, number._number->i);
	*this = op1 * op2;
}

void matrixNumber::operator/= (const matrixNumber& number)
{
	std::complex<double> op1(_number->r, _number->i);
	std::complex<double> op2(number._number->r, number._number->i);
	*this = op1 / op2;
}

int matrixNumber::operator== (const matrixNumber& number) const
{
	if ((number._number->r == _number->r) && (number._number->i == _number->i))
		return 1;
	else
		return 0;
}

int matrixNumber::operator!= (const matrixNumber& number) const
{
	if ((number._number->r == _number->r) && (number._number->i == _number->i))
		return 0;
	else
		return 1;
}

void matrixNumber::operator+= (const std::complex<double>& number)
{
	_number->r += std::real(number);
	_number->i += std::imag(number);
}

void matrixNumber::operator-= (const std::complex<double>& number)
{
	_number->r -= std::real(number);
	_number->i -= std::imag(number);
}

void matrixNumber::operator*= (const std::complex<double>& number)
{
	std::complex<double> op1(_number->r, _number->i);
	*this = op1 * number;
}

void matrixNumber::operator/= (const std::complex<double>& number)
{
	std::complex<double> op1(_number->r, _number->i);
	*this = op1 / number;
}

int matrixNumber::operator== (const std::complex<double>& number) const
{
	if ((std::real(number) == _number->r) && (std::imag(number) == _number->i))
		return 1;
	else
		return 0;
}

int matrixNumber::operator!= (const std::complex<double>& number) const
{
	if ((std::real(number) == _number->r) && (std::imag(number) == _number->i))
		return 0;
	else
		return 1;
}

std::complex<double> matrixNumber::operator+ (const matrixNumber& number) const
{
	std::complex<double> output(_number->r, _number->i);
	return output + std::complex<double>(number._number->r, number._number->i);
}

std::complex<double> matrixNumber::operator- (const matrixNumber& number) const
{
	std::complex<double> output(_number->r, _number->i);
	return output - std::complex<double>(number._number->r, number._number->i);
}

std::complex<double> matrixNumber::operator* (const matrixNumber& number) const
{
	std::complex<double> output(_number->r, _number->i);
	return output * std::complex<double>(number._number->r, number._number->i);
}

std::complex<double> matrixNumber::operator/ (const matrixNumber& number) const
{
	std::complex<double> output(_number->r, _number->i);
	return output / std::complex<double>(number._number->r, number._number->i);
}

std::complex<double> matrixNumber::operator+ (const std::complex<double>& number) const
{
	std::complex<double> output(_number->r, _number->i);
	return output + number;
}

std::complex<double> matrixNumber::operator- (const std::complex<double>& number) const
{
	std::complex<double> output(_number->r, _number->i);
	return output - number;
}

std::complex<double> matrixNumber::operator* (const std::complex<double>& number) const
{
	std::complex<double> output(_number->r, _number->i);
	return output * number;
}

std::complex<double> matrixNumber::operator/ (const std::complex<double>& number) const
{
	std::complex<double> output(_number->r, _number->i);
	return output / number;
}

std::complex<double> matrixNumber::operator+ (const double& number) const
{
	std::complex<double> output(_number->r, _number->i);
	return output + number;
}

std::complex<double> matrixNumber::operator- (const double& number) const
{
	std::complex<double> output(_number->r, _number->i);
	return output - number;
}

std::complex<double> matrixNumber::operator* (const double& number) const
{
	std::complex<double> output(_number->r, _number->i);
	return output * number;
}

std::complex<double> matrixNumber::operator/ (const double& number) const
{
	std::complex<double> output(_number->r, _number->i);
	return output / number;
}

std::ostream& matrixNumber::output(std::ostream& os) const
{
	os << "(" << _number->r << "," << _number->i << ")";
	return os;
}

std::complex<double> matrixNumber::conj() const
{
	return std::complex<double>(_number->r, -_number->i);
}

double matrixNumber::abs() const
{
	return std::sqrt(_number->r * _number->r + _number->i * _number->i);
}

double matrixNumber::real() const
{
	return _number->r;
}

double matrixNumber::imag() const
{
	return _number->i;
}
#endif
